
from perfmetrics import Metric
from perfmetrics import set_statsd_client


@Metric(rate=0.001)
def myfunction():
    """Do something that might be expensive"""


class MyClass(object):
    @Metric(rate=0.001, method=True)
    def mymethod(self):
        """Do some other possibly expensive thing"""


try:
    xrange
except NameError:
    xrange = range  # Python 3


set_statsd_client('statsd://localhost:8125')
for i in xrange(1000000):
    myfunction()
    MyClass().mymethod()



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

class Test_make_statsd_app(unittest.TestCase):

    def setUp(self):
        from perfmetrics import set_statsd_client, statsd_client_stack
        set_statsd_client(None)
        statsd_client_stack.clear()

    def _call(self, nextapp, statsd_uri):
        from perfmetrics import make_statsd_app
        app = make_statsd_app(nextapp, None, statsd_uri)
        if hasattr(app, 'statsd_client'):
            self.addCleanup(app.statsd_client.close)
        return app

    def test_without_statsd_uri(self):
        clients = []

        def dummy_app(_environ, _start_response):
            from perfmetrics import statsd_client
            clients.append(statsd_client())
            return ['ok.']

        app = self._call(dummy_app, '')
        response = app({}, None)
        self.assertEqual(response, ['ok.'])
        self.assertEqual(len(clients), 1)
        self.assertIsNone(clients[0])

    def test_with_statsd_uri(self):
        clients = []

        def dummy_app(_environ, _start_response):
            from perfmetrics import statsd_client
            clients.append(statsd_client())
            return ['ok.']

        app = self._call(dummy_app, 'statsd://localhost:9999')
        response = app({}, None)
        self.assertEqual(response, ['ok.'])
        self.assertEqual(len(clients), 1)
        from perfmetrics.statsd import StatsdClient
        self.assertIsInstance(clients[0], StatsdClient)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from zope.interface import verify

from hamcrest import assert_that

from perfmetrics.interfaces import IStatsdClient

from . import validly_provides
from . import is_true
from . import implements


# pylint:disable=protected-access
# pylint:disable=too-many-public-methods

class MockSocket(object):
    def __init__(self, error=None):
        self.sent = []
        self.error = error

    def sendto(self, data, addr):
        if self.error is not None:
            raise self.error # pylint:disable=raising-bad-type
        self.sent.append((data, addr))

    def close(self):
        pass


class TestBasics(unittest.TestCase):

    @property
    def _class(self):
        from perfmetrics.statsd import StatsdClient
        return StatsdClient

    def _makeOne(self, *args, **kwargs):
        kind = kwargs.pop('kind', None) or self._class
        inst = kind(*args, **kwargs)
        self.addCleanup(inst.close)
        return inst

    def test_provides(self):
        assert_that(self._makeOne(), validly_provides(IStatsdClient))

    def test_implements(self):
        kind = type(self._makeOne())
        assert_that(kind, implements(IStatsdClient))
        # That just checks the declaration of the interface, it doesn't
        # do a static check of attributes.
        assert_that(
            verify.verifyClass(IStatsdClient, kind),
            is_true())

    def test_true(self):
        assert_that(self._makeOne(), is_true())

class TestNullStatsdClient(TestBasics):

    @property
    def _class(self):
        from perfmetrics.statsd import NullStatsdClient
        return NullStatsdClient

class TestStatsdClient(TestBasics):

    sent = ()

    # The main stat name, as entered by the application
    STAT_NAME = 'some.thing'
    # The main stat name, as sent on the wire
    STAT_NAMEB = b'some.thing'
    STAT_NAME2 = 'other.thing'
    STAT_NAME2B = b'other.thing'

    def _make(self, patch_socket=True, error=None, prefix=''):
        obj = self._makeOne(prefix=prefix)

        if patch_socket:
            obj.udp_sock.close()
            obj.udp_sock = MockSocket(error)
            self.sent = obj.udp_sock.sent

        self.addCleanup(obj.close)
        return obj

    def test_ctor_with_defaults(self):
        obj = self._make(patch_socket=False)
        self.assertIsNotNone(obj.udp_sock)
        self.assertIsNotNone(obj.addr)
        self.assertEqual(obj.prefix, '')

    def test_ctor_with_options(self):
        obj = self._make(patch_socket=False, prefix='foo')
        self.assertIsNotNone(obj.udp_sock)
        self.assertIsNotNone(obj.addr)
        self.assertEqual(obj.prefix, 'foo.')

    def test_timing_with_rate_1(self):
        obj = self._make()
        obj.timing(self.STAT_NAME, 750)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':750|ms', obj.addr)])

    def test_timing_with_rate_too_low(self):
        obj = self._make()
        obj.timing(self.STAT_NAME, 750, rate=-1)
        self.assertEqual(self.sent, [])

    def test_timing_with_buf(self):
        obj = self._make()
        buf = []
        obj.timing(self.STAT_NAME, 750, buf=buf)
        self.assertEqual(self.sent, [])
        obj.sendbuf(buf)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':750|ms', obj.addr)])

    def test_gauge_with_rate_1(self):
        obj = self._make()
        obj.gauge(self.STAT_NAME, 50)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':50|g', obj.addr)])

    def test_gauge_with_rate_too_low(self):
        obj = self._make()
        obj.gauge(self.STAT_NAME, 50, rate=-1)
        self.assertEqual(self.sent, [])

    def test_gauge_with_buf(self):
        obj = self._make()
        buf = []
        obj.gauge(self.STAT_NAME, 50, buf=buf)
        self.assertEqual(self.sent, [])
        obj.sendbuf(buf)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':50|g', obj.addr)])

    def test_incr_with_one_metric(self):
        obj = self._make()
        obj.incr(self.STAT_NAME)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':1|c', obj.addr)])

    def test_incr_with_two_metrics(self):
        obj = self._make()
        buf = []
        obj.incr(self.STAT_NAME, buf=buf)
        obj.incr(self.STAT_NAME2, buf=buf)
        obj.sendbuf(buf)
        self.assertEqual(self.sent,
                         [(self.STAT_NAMEB + b':1|c\n' + self.STAT_NAME2B + b':1|c', obj.addr)])

    def test_incr_with_rate_too_low(self):
        obj = self._make()
        obj.incr(self.STAT_NAME, rate=-1)
        self.assertEqual(self.sent, [])

    def test_decr(self):
        obj = self._make()
        obj.decr(self.STAT_NAME)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':-1|c', obj.addr)])

    def test_incr_by_amount_with_one_metric(self):
        obj = self._make()
        obj.incr(self.STAT_NAME, 51)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':51|c', obj.addr)])

    def test_incr_by_amount_with_two_metrics(self):
        obj = self._make()
        buf = []
        obj.incr(self.STAT_NAME, 42, buf=buf)
        obj.incr(self.STAT_NAME2, -41, buf=buf)
        obj.sendbuf(buf)
        self.assertEqual(self.sent,
                         [(self.STAT_NAMEB + b':42|c\n' + self.STAT_NAME2B + b':-41|c', obj.addr)])

    def test_incr_with_rate_hit(self):
        obj = self._make()
        obj.random = lambda: 0.01
        obj.incr(self.STAT_NAME, 51, rate=0.1)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':51|c|@0.1', obj.addr)])

    def test_incr_with_rate_miss(self):
        obj = self._make()
        obj.random = lambda: 0.99
        obj.incr(self.STAT_NAME, 51, rate=0.1)
        self.assertEqual(self.sent, [])

    def test_send_with_ioerror(self):
        obj = self._make(error=IOError('synthetic'))
        obj._send('some.thing:41|g')
        self.assertEqual(self.sent, [])

    def test_sendbuf_with_ioerror(self):
        obj = self._make(error=IOError('synthetic'))
        obj.sendbuf(['some.thing:41|g'])
        self.assertEqual(self.sent, [])

    def test_sendbuf_with_empty_buf(self):
        obj = self._make()
        obj.sendbuf([])
        self.assertEqual(self.sent, [])

    def test_set_add(self):
        obj = self._make()
        obj.set_add(self.STAT_NAME, 42)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':42|s', obj.addr)])

    def test_set_add_with_buf(self):
        obj = self._make()
        buf = []
        obj.set_add(self.STAT_NAME, 42, buf=buf)
        obj.set_add(self.STAT_NAME2, 23, buf=buf)
        obj.sendbuf(buf)
        self.assertEqual(self.sent,
                         [(self.STAT_NAMEB + b':42|s\n' + self.STAT_NAME2B + b':23|s', obj.addr)])


    def test_set_add_with_rate_hit(self):
        obj = self._make()
        obj.random = lambda: 0.01
        obj.set_add(self.STAT_NAME, 51, rate=0.1)
        self.assertEqual(self.sent, [(self.STAT_NAMEB + b':51|s', obj.addr)])

    def test_set_add_with_rate_miss(self):
        obj = self._make()
        obj.random = lambda: 0.99
        obj.set_add(self.STAT_NAME, 51, rate=0.1)
        self.assertEqual(self.sent, [])

class TestStatsdClientMod(TestStatsdClient):

    STAT_NAMEB = b'wrap.some.thing'
    STAT_NAME2B = b'wrap.other.thing'

    @property
    def _class(self):
        from perfmetrics.statsd import StatsdClientMod
        return StatsdClientMod

    def _makeOne(self, *args, **kwargs):
        from perfmetrics.statsd import StatsdClient
        kwargs['kind'] = super(TestStatsdClientMod, self)._class
        wrapped = super(TestStatsdClientMod, self)._makeOne(*args, **kwargs)
        assert type(wrapped) is StatsdClient # pylint:disable=unidiomatic-typecheck
        # Be sure to use a format string that alters the stat name so we
        # can prove the method is getting called. With __getattr__ there, we could
        # silently call through to the wrapped class without knowing it.
        return self._class(wrapped, 'wrap.%s')



# cython: auto_pickle=False,embedsignature=True,always_allow_keywords=False
# -*- coding: utf-8 -*-
"""
Implementation of metrics.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from time import time
from types import MethodType
from weakref import WeakKeyDictionary
import functools
import random as stdrandom

from .clientstack import statsd_client
from .clientstack import client_stack as statsd_client_stack
from .statsd import StatsdClientMod
from .statsd import null_client

logger = __import__('logging').getLogger(__name__)

class _MethodLikeMixin(object):
    __slots__ = ()
    # We may be wrapped by another decorator,
    # so we can't count on __get__ being called.
    # But if it is, we need to act like a bound method.
    #
    # When we compile with Cython, we can't dynamically choose
    # a __get__ impl; the last one defined wins, so we must take the conditional
    # inside the method.
    def __get__(self, inst, klass):
        if inst is None:
            return self
        # Python 2 takes 3 arguments, Python 3 just two. Actually you
        # can get away with passing just the first two to Python 2,
        # but you get '<bound method ?.foo>' and possibly other issues
        # (im_class is None), so it's best to pass all three.
        return MethodType(self, inst, klass) if str is bytes else MethodType(self, inst)

class _AbstractMetricImpl(_MethodLikeMixin):
    __slots__ = (
        'f',
        'random',
        'metric_timing',
        'metric_count',
        'metric_rate',
        'timing_format',
        '__wrapped__',
        '__dict__',
    )
    stat_name = None
    def __init__(self, f, timing, count, rate, timing_format, random):
        self.__wrapped__ = None
        self.f = f
        self.metric_timing = timing
        self.metric_count = count
        self.metric_rate = rate
        self.timing_format = timing_format
        self.random = random

    def __call__(self, *args, **kwargs):
        if self.metric_rate < 1 and self.random() >= self.metric_rate:
            # Ignore this sample.
            return self.f(*args, **kwargs)

        client = statsd_client()

        if client is None:
            # No statsd client has been configured.
            return self.f(*args, **kwargs)

        stat = self.stat_name or self._compute_stat(args)
        # TODO: A lot of this is duplicated with __exit__.
        # Can we do better?
        if self.metric_timing:
            if self.metric_count:
                buf = []
                client.incr(stat, 1, self.metric_rate, buf=buf, rate_applied=True)
            else:
                buf = None

            start = time()

            try:
                return self.f(*args, **kwargs)
            finally:
                end = time()
                elapsed_ms = int((end - start) * 1000.0)
                client.timing(self.timing_format % stat, elapsed_ms,
                              self.metric_rate, buf=buf, rate_applied=True)
                if buf:
                    client.sendbuf(buf)

        else:
            if self.metric_count:
                client.incr(stat, 1, self.metric_rate, rate_applied=True)
            return self.f(*args, **kwargs)

    def _compute_stat(self, args):
        raise NotImplementedError

class _GivenStatMetricImpl(_AbstractMetricImpl):
    __slots__ = (
        'stat_name',
    )
    def __init__(self, stat_name, *args):
        self.stat_name = stat_name
        super(_GivenStatMetricImpl, self).__init__(*args)

    def _compute_stat(self, args): # pragma: no cover
        return self.stat_name

class _MethodMetricImpl(_AbstractMetricImpl):
    __slots__ = (
        'klass_dict',
    )

    def __init__(self, *args):
        self.klass_dict = WeakKeyDictionary()
        super(_MethodMetricImpl, self).__init__(*args)

    def _compute_stat(self, args):
        klass = args[0].__class__
        try:
            stat_name = self.klass_dict[klass]
        except KeyError:
            stat_name = '%s.%s.%s' % (klass.__module__, klass.__name__, self.f.__name__)
            self.klass_dict[klass] = stat_name
        return stat_name


class Metric(object):
    """
    Metric(stat=None, rate=1, method=False, count=True, timing=True)

    A decorator or context manager with options.

    ``stat`` is the name of the metric to send; set it to None to use
    the name of the function or method. ``rate`` lets you reduce the
    number of packets sent to Statsd by selecting a random sample; for
    example, set it to 0.1 to send one tenth of the packets. If the
    ``method`` parameter is true, the default metric name is based on
    the method's class name rather than the module name. Setting
    ``count`` to False disables the counter statistics sent to Statsd.
    Setting ``timing`` to False disables the timing statistics sent to
    Statsd.

    Sample use as a decorator::

        @Metric('frequent_func', rate=0.1, timing=False)
        def frequent_func():
            "Do something fast and frequently."

    Sample use as a context manager::

        def do_something():
            with Metric('doing_something'):
                pass

    If perfmetrics sends packets too frequently, UDP packets may be lost
    and the application performance may be affected.  You can reduce
    the number of packets and the CPU overhead using the ``Metric``
    decorator with options instead of `metric` or `metricmethod`.
    The decorator example above uses a sample rate and a static metric name.
    It also disables the collection of timing information.

    When using Metric as a context manager, you must provide the
    ``stat`` parameter or nothing will be recorded.

    .. versionchanged:: 3.0

        When used as a decorator, set ``__wrapped__`` on the returned object, even
        on Python 2.

    .. versionchanged:: 3.0

        When used as a decorator, the returned object
        has ``metric_timing``, ``metric_count`` and ``metric_rate``
        attributes that can be changed to alter its behaviour.

    """

    def __init__(self, stat=None, rate=1, method=False,
                 count=True, timing=True, timing_format='%s.t',
                 random=stdrandom.random):  # testing hook
        self.stat = stat
        self.rate = rate
        self.method = method
        self.count = count
        self.timing = timing
        self.timing_format = timing_format
        self.random = random
        self.start = 0.0

    def __call__(self, f):
        """
        Decorate a function or method so it can send statistics to statsd.
        """
        func_name = f.__name__
        func_full_name = '%s.%s' % (f.__module__, func_name)

        if self.method:
            metric = _MethodMetricImpl(f, self.timing, self.count,
                                       self.rate, self.timing_format,
                                       self.random)
        else:
            metric = _GivenStatMetricImpl(
                self.stat or func_full_name,
                f, self.timing, self.count,
                self.rate, self.timing_format,
                self.random)

        metric = functools.update_wrapper(metric, f)
        metric.__wrapped__ = f # Python 2 doesn't set this, but it's handy to have.
        return metric

    # Metric can also be used as a context manager.

    def __enter__(self):
        self.start = time()

    def __exit__(self, _typ, _value, _tb):
        rate = self.rate
        if rate < 1 and self.random() >= rate:
            # Ignore this sample.
            return

        client = statsd_client_stack.get()
        if client is not None:
            buf = []
            stat = self.stat
            if stat:
                if self.count:
                    client.incr(stat, rate=rate, buf=buf, rate_applied=True)
                if self.timing:
                    elapsed = int((time() - self.start) * 1000.0)
                    client.timing(self.timing_format % stat, elapsed,
                                  rate=rate, buf=buf, rate_applied=True)
                if buf:
                    client.sendbuf(buf)

class MetricMod(object):
    """Decorator/context manager that modifies the name of metrics in context.

    format is a format string such as 'XYZ.%s'.
    """

    def __init__(self, format):
        self.format = format

    def __call__(self, f):
        """Decorate a function or method to add a metric prefix in context.
        """

        @functools.wraps(f)
        def call_with_mod(*args, **kw):
            client = statsd_client_stack.get()
            if client is None:
                # Statsd is not configured.
                return f(*args, **kw)

            statsd_client_stack.push(StatsdClientMod(client, self.format))
            try:
                return f(*args, **kw)
            finally:
                statsd_client_stack.pop()

        call_with_mod.__wrapped__ = f
        return call_with_mod

    def __enter__(self):
        client = statsd_client_stack.get()

        if client is None:
            statsd_client_stack.push(null_client)
        else:
            statsd_client_stack.push(StatsdClientMod(client, self.format))

    def __exit__(self, _typ, _value, _tb):
        statsd_client_stack.pop()

# pylint:disable=wrong-import-position,wrong-import-order
from perfmetrics._util import import_c_accel
import_c_accel(globals(), 'perfmetrics._metric')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"


from perfmetrics.statsd import StatsdClient

from .observation import Observation


class _TrackingSocket(object):
    """
    Maintain a list of sent packet/address pairs, as well
    as parsed observations/address pairs.
    """
    def __init__(self):
        self.sent_packets = [] # type: List[Tuple[str, bytes]
        self.observations = [] # type: List[Tuple[Observation, bytes]

    def clear(self):
        del self.sent_packets[:]
        del self.observations[:]

    close = clear

    def sendto(self, data, addr):
        # The client encoded to bytes
        assert isinstance(data, bytes)
        # We always want native strings here, that's what the
        # user specified when calling the StatsdClient methods.
        data = data.decode('utf-8') if bytes is not str else data
        self.sent_packets.append((data, addr,))
        for m in Observation.make_all(data):
            self.observations.append((m, addr,))


class FakeStatsDClient(StatsdClient):
    """
    A mock statsd client that tracks sent statsd metrics in memory
    rather than pushing them over a socket. This class is a drop
    in replacement for `perfmetrics.statsd.StatsdClient` that collects statsd
    packets and `~.Observation` that are sent through the client.

    .. versionchanged:: 3.1.0
       Like the normal clients, this object is now always true, whether or
       not any observations have been sent.
    """

    def __init__(self, prefix=''):
        """
        Create a mock statsd client with the given prefix.
        """
        super(FakeStatsDClient, self).__init__(prefix=prefix)

        # Monkey patch the socket to track things in memory instead
        self.udp_sock.close()
        self.udp_sock = _TrackingSocket()

    def clear(self):
        """
        Clears the statsd metrics that have been collected
        """
        self.udp_sock.clear()

    def __bool__(self):
        return True

    __nonzero__ = __bool__ # Python 2

    def __len__(self):
        """
        The number of metrics sent. This accounts for multi metric packets
        that may be sent.
        """
        return len(self.udp_sock.observations)

    def __iter__(self):
        """
        Iterates the `Observations <~.Observation>` provided to this statsd
        client.
        """
        for metric, _ in self.udp_sock.observations:
            yield metric
    iter_observations = __iter__

    def iter_packets(self):
        """
        Iterates the raw statsd packets provided to the statsd client.

        :return: Iterator of native strings.
        """
        for data, _ in self.udp_sock.sent_packets:
            yield data

    @property
    def observations(self):
        """
        A list of `~.Observation` objects collected by this client.

        .. seealso:: `iter_observations`
        """
        return list(self)

    @property
    def packets(self):
        """
        A list of raw statsd packets collected by this client.

        .. seealso:: `iter_packets`
        """
        return list(self.iter_packets())



# -*- coding: utf-8 -*-
"""
perfmetrics is a library for sending software performance metrics
to statsd.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os


from .clientstack import statsd_client
from .clientstack import set_statsd_client
from .clientstack import client_stack as statsd_client_stack


from .statsd import statsd_client_from_uri
from .statsd import StatsdClient

from .metric import Metric
from .metric import MetricMod

from .pyramid import includeme
from .pyramid import tween
from .wsgi import make_statsd_app

__all__ = [
    'metric',
    'metricmethod',
    'Metric',
    'MetricMod',
    'StatsdClient',
    'set_statsd_client',
    'statsd_client',
    'statsd_client_from_uri',
    'statsd_client_stack',
    # Pyramid
    'includeme',
    'tween',
    # WSGI
    'make_statsd_app',
]

class _DisabledMetricDecorator(Metric):

    def __call__(self, f):
        return f

class _DisabledMetricModDecorator(MetricMod):

    def __call__(self, f):
        return f

if os.environ.get('PERFMETRICS_DISABLE_DECORATOR'):
    Metric = _DisabledMetricDecorator
    MetricMod = _DisabledMetricModDecorator

#: @metric:
metric = Metric()

#: @metricmethod:
metricmethod = Metric(method=True)


_uri = os.environ.get('STATSD_URI')
if _uri:
    set_statsd_client(_uri)  # pragma: no cover



# -*- coding: utf-8 -*-
"""
Interfaces for perfmetrics.

If zope.interface is not installed, this file is merely documentation.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint:disable=no-self-argument,no-method-argument

try:
    from zope import interface
except ImportError: # pragma: no cover
    class Interface(object):
        pass
    class Attribute(object):
        def __init__(self, descr):
            self.descr = descr
    class implementer(object):
        def __init__(self, *ifaces):
            pass
        def __call__(self, cls):
            return cls
else:
    Interface = interface.Interface
    Attribute = interface.Attribute
    implementer = interface.implementer


class IStatsdClient(Interface):
    """
    Interface to communicate with a StatsD server.

    Most of the methods below have optional ``rate``, ``rate_applied``,
    and ``buf`` parameters.  The ``rate`` parameter, when set to a value
    less than 1, causes StatsdClient to send a random sample of packets rather
    than every packet.  The ``rate_applied`` parameter, if true, informs
    ``StatsdClient`` that the sample rate has already been applied and the
    packet should be sent regardless of the specified sample rate.

    If the ``buf`` parameter is a list, StatsdClient
    appends the packet contents to the ``buf`` list rather than send the
    packet, making it possible to send multiple updates in a single packet.
    Keep in mind that the size of UDP packets is limited (the limit varies
    by the network, but 1000 bytes is usually a good guess) and any extra
    bytes will be ignored silently.

    """
    def close():
        """
        Release resources (sockets) held by this object.

        .. versionadded:: 3.0
        """

    def timing(stat, value, rate=1, buf=None, rate_applied=False):
        """
        Log timing information in milliseconds.

        *stat* is the name of the metric to record and *value* is
        the timing measurement in milliseconds. Note that Statsd
        maintains several data points for each timing metric, so
        timing metrics can take more disk space than counters or
        gauges.
        """

    def gauge(stat, value, rate=1, buf=None, rate_applied=False):
        """
        Update a gauge value.

        *stat* is the name of the metric to record and *value* is
        the new gauge value. A gauge represents a persistent value
        such as a pool size. Because gauges from different machines
        often conflict, a suffix is usually applied to gauge names;
        this may be done manually or with `MetricMod`.
        """

    def incr(stat, count=1, rate=1, buf=None, rate_applied=False):
        """
        Increment a counter by *count*.

        Note that Statsd clears all counter values every time it sends
        the metrics to Graphite, which usually happens every 10
        seconds. If you need a persistent value, it may be more
        appropriate to use a gauge instead of a counter.
        """

    def decr(stat, count=1, rate=1, buf=None, rate_applied=False):
        """
        Decrement a counter.

        This is the opposite of :meth:`incr`.
        """

    def set_add(stat, value, rate=1, buf=None, rate_applied=False):
        """
        Add a *value* to the set named by *stat*.

        A StatsD set counts the unique occurrences of events (values)
        between flushes.

        For example, if you wanted to count the number of different
        users logging in to an application within the sampling period,
        you could use something like::

            def on_login(user_id):
                client.set_add("logged_in_users", user_id)

        While this method accepts the *rate* parameter, it may be less
        useful here since the point is to let the StatsD server collect
        unique events automatically, and it can't do that if some events
        are dropped, making it only an estimate.

        .. versionadded:: 3.1.0
        """


    def sendbuf(buf):
        """
        Send a UDP packet containing string lines.

        *buf* is a sequence of strings.
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Support for testing that code is instrumented as expected.
"""
from __future__ import print_function, absolute_import, division
__docformat__ = "restructuredtext en"

__all__ = [
    'FakeStatsDClient',
    'Observation',
    'OBSERVATION_KIND_COUNTER',
    'OBSERVATION_KIND_GAUGE',
    'OBSERVATION_KIND_SET',
    'OBSERVATION_KIND_TIMER',
]

from .client import FakeStatsDClient
from .observation import Observation
from .observation import OBSERVATION_KIND_COUNTER
from .observation import OBSERVATION_KIND_GAUGE
from .observation import OBSERVATION_KIND_SET
from .observation import OBSERVATION_KIND_TIMER



import unittest


class Test_ClientStack(unittest.TestCase):
    @property
    def _class(self):
        from perfmetrics.clientstack import ClientStack
        return ClientStack

    def test_ctor(self):
        obj = self._class()
        self.assertIsNotNone(obj.stack)

    def test_push(self):
        obj = self._class()
        client = object()
        obj.push(client)
        self.assertEqual(obj.stack, [client])

    def test_pop_with_client(self):
        obj = self._class()
        client = object()
        obj.stack.append(client)
        got = obj.pop()
        self.assertIs(got, client)
        self.assertEqual(obj.stack, [])

    def test_pop_without_client(self):
        obj = self._class()
        got = obj.pop()
        self.assertIsNone(got)
        self.assertEqual(obj.stack, [])

    def test_get_without_client(self):
        obj = self._class()
        self.assertIsNone(obj.get())

    def test_get_with_client(self):
        obj = self._class()
        client = object()
        obj.stack.append(client)
        self.assertIs(obj.get(), client)

    def test_clear(self):
        obj = self._class()
        client = object()
        obj.stack.append(client)
        obj.clear()
        self.assertEqual(obj.stack, [])



import os
import sys

from setuptools import setup
from setuptools import find_packages
from setuptools import Extension

PYPY = hasattr(sys, 'pypy_version_info')

def read(fname, here=os.path.dirname(__file__)):
    with open(os.path.join(here, fname)) as f:
        return f.read()

README = read('README.rst')
CHANGES = read('CHANGES.rst')

tests_require = [
    'zope.testrunner',
    # nti.testing > ZODB > zodbpickle.
    # But zodbpickle 2.1.0 doesn't build on 3.10.
    # Still waiting on 2.2 (https://github.com/zopefoundation/zodbpickle/pull/61)
    'nti.testing; python_version < "3.10"',
    # transitive dep of nti.testing, which we don't always have, but need
    # for our emulation
    'zope.schema',
    'pyhamcrest >= 1.10',
    'pyperf',
]

###
# Cython
###

# Based on code from
# http://cython.readthedocs.io/en/latest/src/reference/compilation.html#distributing-cython-modules
def _dummy_cythonize(extensions, **_kwargs):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in ('.pyx', '.py'):
                ext = '.c'
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

try:
    from Cython.Build import cythonize
except ImportError:
    # The .c files had better already exist, as they should in
    # an sdist.
    cythonize = _dummy_cythonize

ext_modules = []

# Modules we want to compile with Cython. These *should* have a parallel
# .pxd file (with a leading _) defining cython attributes.
# They should also have a cython comment at the top giving options,
# and mention that they are compiled with cython on CPython.
# The bottom of the file must call import_c_accel.
# We use the support from Cython 28 to be able to parallel compile
# and cythonize modules to a different name with a leading _.
# This list is derived from the profile of bm_simple_iface
# https://github.com/NextThought/nti.externalization/commit/0bc4733aa8158acd0d23c14de2f9347fb698c040
if not PYPY:
    def _source(m, ext):
        # Yes, always /.
        m = m.replace('.', '/')
        return 'src/perfmetrics/' + m + '.' + ext
    def _py_source(m):
        return _source(m, 'py')
    def _pxd(m):
        return _source(m, 'pxd')
    def _c(m):
        return _source(m, 'c')
    # Each module should list the python name of the
    # modules it cimports from as deps. We'll generate the rest.
    # (Not that this actually appears to do anything right now.)

    for mod_name, deps in (
            ('metric', ()),
    ):
        deps = ([_py_source(mod) for mod in deps]
                + [_pxd(mod) for mod in deps]
                + [_c(mod) for mod in deps])

        source = _py_source(mod_name)
        # 'foo.bar' -> 'foo._bar'
        mod_name_parts = mod_name.rsplit('.', 1)
        mod_name_parts[-1] = '_' + mod_name_parts[-1]
        mod_name = '.'.join(mod_name_parts)


        ext_modules.append(
            Extension(
                'perfmetrics.' + mod_name,
                sources=[source],
                depends=deps,
                define_macros=[('CYTHON_TRACE', '0')],
            ))

    try:
        ext_modules = cythonize(
            ext_modules,
            annotate=True,
            compiler_directives={
                #'linetrace': True,
                'infer_types': True,
                'language_level': '3str',
                'always_allow_keywords': False,
                'nonecheck': False,
            },
        )
    except ValueError:
        # 'invalid literal for int() with base 10: '3str'
        # This is seen when an older version of Cython is installed.
        # It's a bit of a chicken-and-egg, though, because installing
        # from dev-requirements first scans this egg for its requirements
        # before doing any updates.
        import traceback
        traceback.print_exc()
        ext_modules = _dummy_cythonize(ext_modules)

setup(
    name='perfmetrics',
    version='3.2.0',
    author='Shane Hathaway',
    author_email='shane@hathawaymix.org',
    maintainer='Jason Madden',
    maintainer_email='jason@nextthought.com',
    description='Send performance metrics about Python code to Statsd',
    keywords="statsd metrics performance monitoring",
    long_description=README + '\n\n' + CHANGES,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*",
    # Get strings from https://pypi.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "License :: Repoze Public License",
        "Topic :: System :: Monitoring",
    ],
    url="https://github.com/zodb/perfmetrics",
    project_urls={
        'Bug Tracker': 'https://github.com/zodb/perfmetrics/issues',
        'Source Code': 'https://github.com/zodb/perfmetrics/',
        'Documentation': 'https://perfmetrics.readthedocs.io',
    },
    license='BSD-derived (http://www.repoze.org/LICENSE.txt)',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    ext_modules=ext_modules,
    include_package_data=True,
    zip_safe=True,
    tests_require=tests_require,
    install_requires=[
        'setuptools',
    ],
    extras_require={
        'test': tests_require,
        'docs': [
            'Sphinx >= 2.1.2',
            'repoze.sphinx.autointerface',
            'pyhamcrest',
            'sphinx_rtd_theme',
        ],
    },
    entry_points="""\
    [paste.filter_app_factory]
    statsd = perfmetrics:make_statsd_app
    """,
)
